@@ -1,9 +1,8 @@
-from .base import *
-
 from .sklearn_models import *
 from .model_xgboost import XGBoost, XGBoostClassifier, XGBoostRegressor
 from .model_lightgbm import LightGBM, LightGBMClassifier, LightGBMRegressor
 from .model_catboost import CatBoost, CatBoostClassifier, CatBoostRegressor
+from .base import *
 
 all_models = {
         'LightGBM': LightGBM,
@@ -15,5 +14,5 @@ all_models = {
         'ExtraTrees': ExtraTrees,
         'XGBoost': XGBoost,
         'CatBoost': CatBoost,
-        'MLP': MLP,
+        #'MLP': MLP,
         }
\ No newline at end of file
